// L0pht DLL Cache Permission Manager
// Copyright (C) 1999, L0pht Heavy Industries, Inc.
// 
// This source is freeware.
// http://www.l0pht.com

#define UNICODE
#include<windows.h>
#include<stdio.h>
#include<io.h>
#include<conio.h>
#include<aclapi.h>
#include"sub_ntddk.h"
#include"resource.h"

HINSTANCE g_hInstance;
SERVICE_STATUS_HANDLE g_hSvcStatus;

// Returned by NtQueryDirectoryObject
typedef struct _OBJDIR_INFORMATION
{
	UNICODE_STRING ObjectName;
	UNICODE_STRING ObjectTypeName; // e.g. Directory, Device ...
	UCHAR          Data[1];        // variable length
} OBJDIR_INFORMATION, *POBJDIR_INFORMATION;


// Loaded Functions


NTSTATUS (__stdcall *NtOpenSection)(
		OUT PHANDLE  SectionHandle,
		IN ACCESS_MASK  DesiredAccess,
		IN POBJECT_ATTRIBUTES  ObjectAttributes
		);

NTSTATUS (__stdcall *NtOpenDirectoryObject)(
		OUT PHANDLE  DirectoryHandle,
		IN ACCESS_MASK  DesiredAccess,
		IN POBJECT_ATTRIBUTES  ObjectAttributes
		);

VOID (__stdcall *RtlInitUnicodeString)(
		IN OUT PUNICODE_STRING  DestinationString,
		IN PCWSTR  SourceString
		);

ULONG (__stdcall *RtlNtStatusToDosError) (
		IN NTSTATUS Status
		);
VOID  (__stdcall *NtClose)(
		IN HANDLE  Object
		);

NTSTATUS (__stdcall *NtOpenSymbolicLinkObject)(
		OUT PHANDLE SymLinkObjHandle,
		IN ACCESS_MASK DesiredAccess, 
		IN POBJECT_ATTRIBUTES ObjectAttributes);

NTSTATUS (__stdcall *NtQuerySymbolicLinkObject)(
		IN HANDLE SymLinkObjHandle,
		OUT PUNICODE_STRING LinkName, // resolved name of link
		OUT PULONG DataWritten); // can be NULL


// Query information about the content of a directory object
//
//   DirObjInformation   buffer must be large enough to hold
//                       the name strings too
//
//   GetNextIndex=TRUE   return the index of the next object
//                       in this directory in ObjectIndex
//   GetNextIndex=FALSE  return the number of objects in
//                       this directory in ObjectIndex
//
//   IgnoreInputIndex=TRUE  ignore input value of ObjectIndex
//                          always start at index 0
//   IgnoreInputIndex=FALSE use input value of ObjectIndex
//
//   ObjectIndex  zero based index of object in the directory
//                depends on GetNextIndex and IgnoreInputIndex

NTSTATUS (__stdcall *NtQueryDirectoryObject)(
		IN HANDLE DirObjHandle,
		OUT POBJDIR_INFORMATION DirObjInformation, 
		IN ULONG BufferLength, // size of info buffer
		IN BOOLEAN GetNextIndex, 
		IN BOOLEAN IgnoreInputIndex, 
		IN OUT PULONG ObjectIndex,
		OUT PULONG DataWritten); // can be NULL


// Close a handle (not specific to Object Manager)


// PrintError
// Formats an NTSTATUS error for output
void PrintError( char *message, NTSTATUS status )
{
	char *errMsg;

	FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL, RtlNtStatusToDosError( status ), 
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
			(LPTSTR) &errMsg, 0, NULL );
	printf("%s: %S\n", message, errMsg );
	LocalFree( errMsg );
}


// Opens a Section
HANDLE OpenSection(WCHAR *sectionName)
{
	NTSTATUS		status;
	HANDLE			section;
	UNICODE_STRING	sectionString;
	OBJECT_ATTRIBUTES attributes;

	RtlInitUnicodeString( &sectionString, sectionName );	

	InitializeObjectAttributes( &attributes, &sectionString,
								OBJ_CASE_INSENSITIVE, NULL, NULL );			
	status = NtOpenSection( &section, SECTION_ALL_ACCESS, &attributes );

	if( !NT_SUCCESS( status )) {
		PrintError( "Could not open dll section!", status );
		return NULL;
	}

	return section;
}


// LocateEntryPoints
BOOL LocateEntryPoints()
{
	if( !(*(DWORD *)&RtlInitUnicodeString = (DWORD) GetProcAddress( GetModuleHandle(TEXT("ntdll.dll")),
			"RtlInitUnicodeString" )) ) return FALSE;
	if( !(*(DWORD *)&NtOpenDirectoryObject = (DWORD) GetProcAddress( GetModuleHandle(TEXT("ntdll.dll")),
			"NtOpenDirectoryObject" )) ) return FALSE;
	if( !(*(DWORD *)&NtOpenSection = (DWORD) GetProcAddress( GetModuleHandle(TEXT("ntdll.dll")),
			"NtOpenSection" )) ) return FALSE;
	if( !(*(DWORD *)&RtlNtStatusToDosError = (DWORD) GetProcAddress( GetModuleHandle(TEXT("ntdll.dll")),
			"RtlNtStatusToDosError" )) ) return FALSE;
	if( !(*(DWORD *)&NtClose = (DWORD) GetProcAddress( GetModuleHandle(TEXT("ntdll.dll")),
			"NtClose" )) ) return FALSE;
	if( !(*(DWORD *)&NtQueryDirectoryObject = (DWORD) GetProcAddress( GetModuleHandle(TEXT("ntdll.dll")),
			"NtQueryDirectoryObject" )) ) return FALSE;
	
	return TRUE;
}

BOOL CALLBACK DlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	static BOOL *bActive;
	static BOOL bNewActive;
	WORD wNotifyCode,wID;
	HWND hwndCtl;

	switch(uMsg) {
	case WM_INITDIALOG:
		bActive=(BOOL *) lParam;
		if(*bActive) {
			EnableWindow(GetDlgItem(hwndDlg,IDC_INSTALL),FALSE);
			SendDlgItemMessage(hwndDlg,IDC_UNINSTALL,BM_SETCHECK,BST_CHECKED,0);
			bNewActive=FALSE;
		} else {
			EnableWindow(GetDlgItem(hwndDlg,IDC_UNINSTALL),FALSE);
			SendDlgItemMessage(hwndDlg,IDC_INSTALL,BM_SETCHECK,BST_CHECKED,0);
			bNewActive=TRUE;
		}
		return TRUE;
	case WM_COMMAND:	
		wNotifyCode = HIWORD(wParam); // notification code 
		wID = LOWORD(wParam);         // item, control, or accelerator identifier 
		hwndCtl = (HWND) lParam;      // handle of control  
		switch(wID) {
		case IDOK:
			*bActive=bNewActive;
			EndDialog(hwndDlg,IDOK);
			return TRUE;
		case IDCANCEL:
			EndDialog(hwndDlg,IDCANCEL);
			return TRUE;
		}
	}

	return FALSE;
}

BOOL DoInstaller(void)
{
	
	// Open a handle to the SC Manager database.
	SC_HANDLE  SCManagerHandle;
    SCManagerHandle = OpenSCManager(
                      NULL,                   // local machine           
                      NULL,                   // ServicesActive database 
                      SC_MANAGER_ALL_ACCESS); // full access rights      
    if (SCManagerHandle==NULL) {
		MessageBox(NULL,L"Could not open service control manager.",L"Error.",MB_OK|MB_ICONSTOP|MB_SETFOREGROUND|MB_TOPMOST);
		return -1;
	}

	// Get service
	BOOL bActive;
	SC_HANDLE  SCServiceHandle;
	SCServiceHandle=OpenService(SCManagerHandle,L"DLLFIXSYS",SERVICE_START | SERVICE_STOP | DELETE);
	if (SCServiceHandle == NULL) {
		bActive=FALSE;	
	} else {
		bActive=TRUE;
	}

	// Prompt user
	if(DialogBoxParam(g_hInstance,MAKEINTRESOURCE(IDD_INSTALL),NULL,DlgProc,(LPARAM)&bActive)==IDOK) {
		WCHAR svDestFile[MAX_PATH+1];
		GetSystemDirectory(svDestFile,MAX_PATH);
		lstrcpyn(svDestFile+lstrlen(svDestFile),L"\\DLLFIXSYS.EXE",MAX_PATH-(lstrlen(svDestFile)));

		if(bActive) {
			// Add service
			WCHAR svSourceFile[MAX_PATH+1];
			if(GetModuleFileName(g_hInstance,svSourceFile,MAX_PATH)) {
				WCHAR svDestFile[MAX_PATH+1];
				GetSystemDirectory(svDestFile,MAX_PATH);
				lstrcpyn(svDestFile+lstrlen(svDestFile),L"\\DLLFIXSYS.EXE",MAX_PATH-(lstrlen(svDestFile)));
				CopyFile(svSourceFile,svDestFile,FALSE);

				SCServiceHandle=CreateService(SCManagerHandle,
					L"DLLFIXSYS",
					L"L0pht DLL Cache Permission Manager",
					SERVICE_ALL_ACCESS,
					SERVICE_WIN32_OWN_PROCESS | SERVICE_INTERACTIVE_PROCESS,
					SERVICE_AUTO_START,
					SERVICE_ERROR_NORMAL,
					svDestFile,
					NULL,
					NULL,
					NULL,
					NULL,
					NULL);
				if(SCServiceHandle == NULL) {
					MessageBox(NULL,L"Couldn't install service.",L"Error",MB_OK|MB_ICONSTOP|MB_SETFOREGROUND|MB_TOPMOST);
					CloseServiceHandle(SCManagerHandle);
					return -1;
				}
				MessageBox(NULL,L"Service installed successfully.\nChanges now taking place will occur\nevery time you boot the system.",L"Service Installed",MB_OK|MB_ICONINFORMATION|MB_SETFOREGROUND|MB_TOPMOST);
				
				if(StartService(SCServiceHandle,0,NULL)==FALSE) {
					MessageBox(NULL,L"Couldn't start service.",L"Error",MB_OK|MB_ICONSTOP|MB_SETFOREGROUND|MB_TOPMOST);
					CloseServiceHandle(SCServiceHandle);
					CloseServiceHandle(SCManagerHandle);
					return -1;
				}

				CloseServiceHandle(SCServiceHandle);
				CloseServiceHandle(SCManagerHandle);

				return 0;

			} else {
				MessageBox(NULL,L"Couldn't get source file.",L"Error",MB_OK|MB_ICONSTOP|MB_SETFOREGROUND|MB_TOPMOST);
				CloseServiceHandle(SCManagerHandle);
				return -1;
			}
		} else {
			// Remove service
			if(DeleteService(SCServiceHandle)) {
				MessageBox(NULL,L"Service removed.\nYou should reboot now to restore system settings.\n",L"Success",MB_OK|MB_ICONINFORMATION|MB_SETFOREGROUND|MB_TOPMOST);
				CloseServiceHandle(SCServiceHandle);
				CloseServiceHandle(SCManagerHandle);
				return -1;
			}
			
			MessageBox(NULL,L"Service could not be removed.\n",L"Error",MB_OK|MB_ICONSTOP|MB_SETFOREGROUND|MB_TOPMOST);
			CloseServiceHandle(SCServiceHandle);
			CloseServiceHandle(SCManagerHandle);	
			return -1;
		}
	}

	if(SCServiceHandle) CloseServiceHandle(SCServiceHandle);	
	CloseServiceHandle(SCManagerHandle);	
	
	return -1;
}

int FixDLLs(void)
{
	NTSTATUS stat;

	// Load entry points
	if( !LocateEntryPoints() ) {
		MessageBox(NULL,TEXT("Unable to locate system call entry points.\nEnsure you are running on a Windows NT machine.\n"),TEXT("Error"),MB_OK|MB_ICONSTOP|MB_SETFOREGROUND|MB_TOPMOST);
		return -1;
	}
		
	// Get 'KnownDlls' directory object
		
	UNICODE_STRING dirString;
	OBJECT_ATTRIBUTES dirAttr;
	HANDLE hdir;
	
	RtlInitUnicodeString( &dirString, TEXT("\\KnownDlls"));	
	InitializeObjectAttributes( &dirAttr, &dirString, OBJ_CASE_INSENSITIVE, NULL, NULL);			
	
	stat=NtOpenDirectoryObject(&hdir,
		DIRECTORY_QUERY | DIRECTORY_TRAVERSE | WRITE_DAC,
		&dirAttr);
	if(!NT_SUCCESS(stat)) {
		MessageBox(NULL,TEXT("Unable to open 'KnownDlls' directory to modify access control.\n"),TEXT("Error"),MB_OK|MB_ICONSTOP|MB_SETFOREGROUND|MB_TOPMOST);
		return -1;
	}
	
	// Create appropriate DACL

	SID worldsid;
	PACL pAcl;
	int acelen;
	
	worldsid.Revision=SID_REVISION;
	worldsid.SubAuthorityCount=0;
	worldsid.IdentifierAuthority.Value[0]=0; 
	worldsid.IdentifierAuthority.Value[1]=0; 
	worldsid.IdentifierAuthority.Value[2]=0; 
	worldsid.IdentifierAuthority.Value[3]=0; 
	worldsid.IdentifierAuthority.Value[4]=0; 
	worldsid.IdentifierAuthority.Value[5]=1; 
	
	acelen=sizeof (ACCESS_ALLOWED_ACE)- sizeof (DWORD) + GetLengthSid (&worldsid);
	pAcl=(PACL) malloc(acelen+sizeof(ACL));
	if(pAcl==NULL) {
		MessageBox(NULL,TEXT("Unable to allocate memory.\n"),TEXT("Error"),MB_OK|MB_ICONSTOP|MB_SETFOREGROUND|MB_TOPMOST);
		NtClose(hdir);
		return -1;
	}
	InitializeAcl(pAcl,acelen+sizeof(ACL),ACL_REVISION);

	AddAccessAllowedAce(pAcl,ACL_REVISION, READ_CONTROL|DIRECTORY_QUERY|DIRECTORY_TRAVERSE, &worldsid);
	
	// Change directory permissions

	if(SetSecurityInfo(hdir,SE_KERNEL_OBJECT,DACL_SECURITY_INFORMATION,NULL,NULL,pAcl,NULL)!=ERROR_SUCCESS) {
		MessageBox(NULL,TEXT("Unable to set security on KnownDlls directory.\n"),TEXT("Error"),MB_OK|MB_ICONSTOP|MB_SETFOREGROUND|MB_TOPMOST);
		free(pAcl);
		NtClose(hdir);
		return -1;
	}

	// Now go through all the file mapping objects (sections) and fix them up too.
	
	InitializeAcl(pAcl,acelen+sizeof(ACL),ACL_REVISION);
	AddAccessAllowedAce(pAcl, ACL_REVISION, SECTION_MAP_READ | SECTION_MAP_WRITE | SECTION_MAP_EXECUTE | SECTION_QUERY | READ_CONTROL, &worldsid);

	POBJDIR_INFORMATION podi;
	DWORD index;
	podi=(POBJDIR_INFORMATION) malloc(2048); // arbitrary, but good enough
	index=0;
	while(NT_SUCCESS(NtQueryDirectoryObject(hdir,podi,2048,TRUE,FALSE,&index,NULL))) {
		if(wcscmp(podi->ObjectTypeName.Buffer,L"Section")==0) {
			HANDLE hsec;
			WCHAR svSection[1024];
			
			wsprintf(svSection,L"\\KnownDlls\\%.260s",podi->ObjectName.Buffer);
			hsec=OpenSection(svSection);
			if(hsec) {
				if(SetSecurityInfo(hsec,SE_KERNEL_OBJECT,DACL_SECURITY_INFORMATION,NULL,NULL,pAcl,NULL)!=ERROR_SUCCESS) {
					MessageBox(NULL,TEXT("Unable to set security on DLL section.\n"),podi->ObjectTypeName.Buffer,MB_OK|MB_ICONSTOP|MB_SETFOREGROUND|MB_TOPMOST);
				}
				NtClose(hsec);
			}
		}				
	}
	free(podi);
	free(pAcl);
	NtClose(hdir);
	
	return 0;
}

VOID WINAPI SvcHandler(DWORD fdwControl)
{
	SERVICE_STATUS stat;

	stat.dwServiceType=SERVICE_WIN32_OWN_PROCESS | SERVICE_INTERACTIVE_PROCESS;
	stat.dwCurrentState=SERVICE_RUNNING;
	stat.dwControlsAccepted=0;
	stat.dwWin32ExitCode=NO_ERROR;
	stat.dwCheckPoint=0;
	stat.dwWaitHint=500;

	switch(fdwControl) {
	case SERVICE_CONTROL_INTERROGATE:
		SetServiceStatus(g_hSvcStatus,&stat);
		break;
	}
}

VOID WINAPI ServiceMain(DWORD dwArgc, LPTSTR *lpszArgv)
{
	g_hSvcStatus=RegisterServiceCtrlHandler(L"DLLFIXSYS",SvcHandler);

	SERVICE_STATUS stat;

	stat.dwServiceType=SERVICE_WIN32_OWN_PROCESS | SERVICE_INTERACTIVE_PROCESS;
	stat.dwCurrentState=SERVICE_RUNNING;
	stat.dwControlsAccepted=0;
	stat.dwWin32ExitCode=NO_ERROR;
	stat.dwCheckPoint=0;
	stat.dwWaitHint=500;

	SetServiceStatus(g_hSvcStatus,&stat);		

	FixDLLs();
}

// Entry Point
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	g_hInstance=hInstance;
	
	WCHAR svModule[MAX_PATH+1];
	GetModuleFileName(hInstance,svModule,MAX_PATH);
	WCHAR *svFilename;
	svFilename=wcsrchr(svModule,'\\')+1;

	// Command line
	if(wcsicmp(svFilename,L"DLLFIXSYS.EXE")!=0) {
		DoInstaller();
		return 0;
	}

	// Register service handler
	SERVICE_TABLE_ENTRY ste;
	ste.lpServiceName=L"DLLFIXSYS";
	ste.lpServiceProc=ServiceMain;
	StartServiceCtrlDispatcher(&ste);
	return 0;
}